// LICENSE:  LGPL2.1
// Copyright (c) 2002 Tomoya Tokairin

#include <windows.h>
#include <stdio.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>
#include "oggdec.h"
#include "common.h"

BOOL ResetWave();


//------------------------------------------------
// sync buffer  data Zbg
DWORD WriteToSyncBuf(HANDLE hInputFile,ogg_sync_state* p_o_sync,DWORD dwReadSize){
	
	char* lpszBuffer;
	DWORD dwByte;
	
	lpszBuffer = ogg_sync_buffer(p_o_sync,dwReadSize); 
	ReadFile(hInputFile,lpszBuffer,dwReadSize, &dwByte, NULL);
	ogg_sync_wrote(p_o_sync,dwByte); 
	
	return dwByte;
}





//----------------------------------------
// page infomation ̎擾
BOOL GetOggPageInfo(LPBYTE buffer,
					DWORD dwBufSize,
					LONG nSerial, 
					LPDWORD lpdwOffset, // offset 
					LPDWORD lpdwPageCount, // page counter
					LPDWORD lpdwAbPcmPos, // absolute PCM position 
					LPDWORD lpdwPacketNum, // packet number
					LPDWORD lpdwPageSize  // page size (byte)
					){

	DWORD i,i2,dwPageSize,dwPacketNum,dwOffset;
	BYTE cSegment,tmp;

	dwPageSize = 27;
	dwOffset = *lpdwOffset;
	for(i=dwOffset;i<dwBufSize-4;i++){
		
		if(((*(LPDWORD)(buffer+i)) ^ OGG_HEAD_DWORD) == 0)
		{

			// serial number
			if(*(LPLONG)(buffer+i+14) == nSerial){
					
				*lpdwOffset = i;
				
				// absolute position 
				memcpy(lpdwAbPcmPos,buffer+i+6,sizeof(DWORD)); 
				
				// page counter
				memcpy(lpdwPageCount,buffer+i+18,sizeof(DWORD)); 
				
				// page_segments
				cSegment = buffer[i+26];
				dwPageSize += cSegment;
				
				// packet number
				dwPacketNum = 0;
				for(i2 = i + 27; i2 < i + 27 + cSegment; i2++){
					tmp = buffer[i2];
					if(tmp < 255) dwPacketNum++;
					dwPageSize += tmp;
				}
				
				*lpdwPacketNum = dwPacketNum;
				*lpdwPageSize = dwPageSize;
				
				return TRUE;
			}
		}
	}


	return FALSE;
}




//------------------------------------------------
// absolute PCM size ̎擾
DWORD GetOggPCMAbPos(HANDLE hInputFile,
					 DWORD dwFileSize,
					 LONG nSerial,
					 LPDWORD lpdwTotalPage,
					 BYTE* buffer,DWORD dwBufSize){

	DWORD i,dwByte,dwCurPos,dwPos;

	*lpdwTotalPage = 0;
	dwPos = -1;

	dwCurPos = SetFilePointer(hInputFile,0, NULL,FILE_CURRENT);
	if(dwFileSize > dwBufSize) SetFilePointer(hInputFile,-(LONG)dwBufSize, NULL,FILE_END);
	else SetFilePointer(hInputFile,0, NULL,FILE_BEGIN);
	ReadFile(hInputFile,buffer,dwBufSize, &dwByte, NULL);
	
	for(i=0;i<dwByte-4;i++){
		
		if(((*(LPDWORD)(buffer+i)) ^ OGG_HEAD_DWORD) == 0)
		{
			// final page = 0x04
			if(buffer[i+5] & 0x04){ 
				
				// serial
				if(*(LPLONG)(buffer+i+14) == nSerial){
					memcpy(&dwPos,buffer+i+6,sizeof(DWORD));   // absolute PCM size
					memcpy(lpdwTotalPage,buffer+i+18,sizeof(DWORD)); // page
					goto L_END;
				}
			}
		}
	}

L_END:

	SetFilePointer(hInputFile,dwCurPos, NULL,FILE_BEGIN);

	return dwPos;
}



//--------------------------------------------------
// seek (ST)
BOOL SeekAllFile(HANDLE hInputFile,
				 DWORD dwSeekPos,
				 DWORD dwReadByte,
				 WAVEFORMATEX waveFmt,
				 LONG nSerial,
				 DWORD dwTargetPos,
				 BYTE* buffer,
				 LARGE_INTEGER* lpliPrePos){

	DWORD dwByte;
	DWORD dwAbPcmPos;
	DWORD dwPagePos;
	DWORD dwPageCount;
	DWORD dwPacketNum;
	DWORD dwPageSize;

	SetFilePointer(hInputFile,dwSeekPos,NULL,FILE_BEGIN);
	ReadFile(hInputFile,buffer,dwReadByte, &dwByte, NULL);
	
	dwPagePos = 0;
	while(1){
		
		if(!GetOggPageInfo(buffer,
			dwByte,
			nSerial,
			&dwPagePos,
			&dwPageCount,
			&dwAbPcmPos,
			&dwPacketNum,
			&dwPageSize)) return FALSE;

#ifdef _DEBUG
		fprintf(stderr,"page: %d,  pos = %u, target = %u \n",dwPageCount,dwAbPcmPos,dwTargetPos);
#endif	
		if(dwAbPcmPos * waveFmt.nBlockAlign >= dwTargetPos) break;
		
		dwPagePos  += dwPageSize;
		if(dwPagePos >= dwByte) return FALSE;

		lpliPrePos->QuadPart = dwAbPcmPos;
	}

	SetFilePointer(hInputFile,dwSeekPos+dwPagePos,NULL,FILE_BEGIN);
	return TRUE;

}




//------------------------------------------------------
// seek (y[WP)
#define SEEK_LOOP_COUNT		3   // Ƃ肠 3 ʌs
BOOL SeekOggPage(HANDLE hInputFile,
				 DWORD dwFileSize,
				 LONG nSerial,
				 LONG nBitrate,
				 double dSeekTime,  // msec
				 WAVEFORMATEX waveFmt,
				 BYTE* buffer,DWORD dwMaxBufSize, // obt@TCY͏\傫Ƃ邱
				 LARGE_INTEGER* lpliPrePos
				 ){
	
	LONG nSeekPos;
	DWORD i,i2,dwByte;
	DWORD dwTargetPos; 
	DWORD dwBufSize;
	DWORD dwEstimatedPageSize;
	
	DWORD dwAbPcmPos[2];
	DWORD dwPagePos[3];
	DWORD dwPageCount[2];
	DWORD dwPacketNum[2];
	DWORD dwPageSize[2];

	dwTargetPos = (DWORD)(waveFmt.nAvgBytesPerSec*dSeekTime/1000.); // ڕWʒu
	
	//----------------------------------------------------------
	// t@CTCYobt@TCY菬ȂST
	if(dwFileSize < dwMaxBufSize){
		nSeekPos = 0;
		dwBufSize = dwFileSize;
		return(SeekAllFile(hInputFile,nSeekPos,dwBufSize,waveFmt,nSerial,dwTargetPos,buffer,lpliPrePos));
	}

	//-------------------------------------------------------------
	// ȉAȂƂt@CTCYobt@TCY傫ꍇ̏

	dwEstimatedPageSize = 64*1024;  // 1 y[W̃TCY͂Ƃ肠 64k ͏
	nSeekPos = max(0,(LONG)(nBitrate/8.*(dSeekTime/1000.)));  // ̂̈ʒu猟n߂
	if((DWORD)nSeekPos > dwFileSize) nSeekPos = dwFileSize - dwMaxBufSize;
	
#ifdef _DEBUG
	fprintf(stderr,"time = %lf sec\ntarget = %u byte\nseek pos = %u byte\nfilesize = %u\n"
		,dSeekTime/1000,dwTargetPos,nSeekPos,dwFileSize);
#endif

	// V[Nʒu
	for(i2=0;i2<SEEK_LOOP_COUNT;i2++){ 
		
		// ǂݍ݃TCY̍XV
		dwBufSize = dwMaxBufSize > 8 * dwEstimatedPageSize
			? 8 * dwEstimatedPageSize : dwMaxBufSize; // ܂ 8 y[W炢}[WƂǂ
		
#ifdef _DEBUG
		fprintf(stderr,"start = %u buf = %u\n",nSeekPos,dwBufSize);
#endif
		
		SetFilePointer(hInputFile,nSeekPos,NULL,FILE_BEGIN);
		ReadFile(hInputFile,buffer,dwBufSize, &dwByte, NULL);
		
		// obt@̐擪琔 2 ̃y[W̏擾
		dwPagePos[0] = dwPagePos[1] = 0;
		for(i=0;i<2;i++){
			
			if(!GetOggPageInfo(buffer,
				dwByte,
				nSerial,
				&dwPagePos[i],
				&dwPageCount[i],
				&dwAbPcmPos[i],
				&dwPacketNum[i],
				&dwPageSize[i]))
			{
				if(dwByte == dwBufSize) return FALSE; // ǂݍ݃G[
				
#ifdef _DEBUG
				fprintf(stderr,"\nfinal page!\n");
#endif
				// t@C̍Ō̕őST
				nSeekPos = dwFileSize - dwMaxBufSize;
				dwBufSize = dwMaxBufSize;
				return(SeekAllFile(hInputFile,nSeekPos,dwBufSize,waveFmt,nSerial,dwTargetPos,buffer,lpliPrePos));

			}
			else
			{
				
				dwAbPcmPos[i] *= waveFmt.nBlockAlign;
				dwPagePos[i+1] = dwPagePos[i] + dwPageSize[i];
				
#ifdef _DEBUG
				fprintf(stderr,"[%d %d]\n",i2,i);
				fprintf(stderr,"pos = %u page = %u ad = %u ",dwPagePos[i],dwPageCount[i],dwAbPcmPos[i]);
				fprintf(stderr,"pagesize = %u packet = %u ",dwPageSize[i],dwPacketNum[i]);
				fprintf(stderr,"sec  = %lf\n",(double)dwAbPcmPos[i]/waveFmt.nAvgBytesPerSec);
#endif
			}
		}
		dwEstimatedPageSize = dwPageSize[1];
		
		// 2 ̃y[W absolute PCM position ̊Ԃ target position 
		// ͂܂Ă OK
		if(dwAbPcmPos[0] <= dwTargetPos && dwTargetPos < dwAbPcmPos[1]) break;

		// ʒu̍XV
		if(i2 < SEEK_LOOP_COUNT-1){
			
			// 2 ̃y[W̃TCY absolute PCM position  target position ̊֌W
			//  seek positon 𐄒
			if(dwTargetPos > dwAbPcmPos[0])
			{
				nSeekPos = nSeekPos +dwPagePos[1] 
				+(LONG)((double)dwPageSize[1]
					*((double)dwTargetPos-(double)dwAbPcmPos[0])
					/((double)dwAbPcmPos[1]-(double)dwAbPcmPos[0]));
			}
			else
			{
				nSeekPos = nSeekPos 
					+(LONG)((double)dwPageSize[1]
					*((double)dwTargetPos-(double)dwAbPcmPos[0])
					/((double)dwAbPcmPos[1]-(double)dwAbPcmPos[0]));
			}
			
			nSeekPos = max(0,nSeekPos);
#ifdef _DEBUG
			fprintf(stderr,"new seek pos = %u, target = %u, filesize = %u\n",nSeekPos,dwTargetPos,dwFileSize);
#endif
			// V[N悪t@CTCY𒴂Ă
			// t@C̍Ō̕őST
			if((DWORD)nSeekPos > dwFileSize){
	
#ifdef _DEBUG
				fprintf(stderr,"\nseek over\n");
#endif
				nSeekPos = dwFileSize - dwMaxBufSize;
				dwBufSize = dwMaxBufSize;
				return(SeekAllFile(hInputFile,nSeekPos,dwBufSize,waveFmt,nSerial,dwTargetPos,buffer,lpliPrePos));
			}
			
		}


	}


#ifdef _DEBUG
	fprintf(stderr,"SeekOggPage: seek to %u ,page %d\n",nSeekPos + dwPagePos[0],dwPageCount[0]);
#endif

	SetFilePointer(hInputFile,nSeekPos + dwPagePos[0],NULL,FILE_BEGIN);
	lpliPrePos->QuadPart = dwAbPcmPos[0]/waveFmt.nBlockAlign;

	return TRUE;
}





//------------------------------------------------------
// SSE  x32767 (FPU  4,5% ʂ͑Ȃ)
void ConvF2S_SSE(float* pcmdata,
					 float* vol,
					 DWORD dwSampleNum){


	_asm{
		mov edi,dwSampleNum
		mov ecx,pcmdata
		shr edi,2 
		mov eax,vol
		inc edi // [
		movaps  xmm1,[eax] // xmm1 = vol

L1:
		movaps	xmm0,[ecx]  // xmm0 = pcm  
		mulps	xmm0,xmm1  // xmm0 *= vol
		movaps [ecx],xmm0  // tmpBuf = xmm0

		add ecx,16		   // ecx += sizeof(float)*4
		dec edi
		jne	L1
	}
	
}
	




//------------------------------------------------------
// float ^ PCM data  short ɒ֐
void ConvFloat2Short(float* pcmdata,
					 short* pcmbuf,
					 float* vol,
					 DWORD dwSampleNum,
					 DWORD dwOffset,
					 DWORD dwChannels,
					 DWORD dwSSE){
	
	LONG nVal;
	DWORD i;

	// x32767 
	if(dwSSE) ConvF2S_SSE(pcmdata,vol,dwSampleNum);
	else
		for(i=0; i< dwSampleNum; i++) pcmdata[i] = (float)((double)pcmdata[i]*32767.);


	for(i=0; i< dwSampleNum; i++){
		
		// float ->  short
		nVal = (LONG)pcmdata[i];
		
		// NbsO΍
		if(nVal > 32767){
#ifdef _DEBUG
			fprintf(stderr,"\nclipping!!\n");
#endif
			nVal = 32767;
		}
		else if(nVal < -32768){
#ifdef _DEBUG
			fprintf(stderr,"\nclipping!!\n");
#endif
			nVal= -32768;
		}
		
		pcmbuf[dwOffset+i*dwChannels] = (short)nVal;
	}

  
}



//-------------------------------------------------------------------
// SSE,SSE2,MMX T|[g
void GetCpuInfo(LPDWORD lpdwSSE,LPDWORD lpdwSSE2,LPDWORD lpdwMMX){

	__asm{

		//  CPUID gȂ CPU ȂĖƎvꉞ
		pushfd
		pop eax
		mov ecx,eax
		xor eax,(1<<21) 
		push eax
		popfd
		pushfd
		pop eax
		cmp eax,ecx 
		je EXIT	// if(eax == ecx) goto EXIT;

		// get CPUID 
		xor eax,eax
		cpuid
		cmp eax,1 
		jb EXIT // if(eax < 1) goto EXIT;
		mov eax,1
		cpuid

		push edx
		push edx
		push edx

		// SSE2 = 26
		pop edx
		and edx,(1<<26) 
		mov ebx,lpdwSSE2
		mov dword ptr [ebx], edx

		// SSE = 25
		pop edx
		and edx,(1<<25) 
		mov ebx,lpdwSSE
		mov dword ptr [ebx], edx

		// MMX = 24
		pop edx
		and edx,(1<<24) 
		mov ebx,lpdwMMX
		mov dword ptr [ebx], edx

EXIT:
	}

}



//--------------------------------------------------------------
// ݂̃NbNTCN擾֐
// ReaD Time Stamp Counter (RDTSC) ߂g
void GetCurCycle(LARGE_INTEGER* lpli){
	
	__asm {
		mov ecx,lpli  // ecx = lpli
		rdtsc  // eax = low part, edx = high part
		mov dword ptr [ecx],eax  // lpli->LowPart = eax
		mov dword ptr [ecx+4],edx // lpli->HighPart = edx
	}

}

// EOF